(in-package :biodb)

;;  ***************************************************************
;;
;;      Copyright 2000 SRI International.  All Rights Reserved.
;;
;;  ***************************************************************


;;;;      BioDB-Loader
;;;;
;;;;  This file contains the main functions for the BioDB-Loader package.



;;;========================================================================
;;; EcoCyc parsing function

(defun parse-ecocyc (filename 
		     &key 
		     index-attributes)
  (let ((separator-alist
	 '((default (#\Newline))
	   ))
	continuator-alist
	removal-alist
	(data-spacers " - ")
	)
    
  
    (parse-flatfile-db filename "EcoCyc" "//"
		       separator-alist continuator-alist
		       removal-alist data-spacers
		       :primary-key 'UNIQUE-ID
		       :index-attributes index-attributes)))

;;; TEST VERSION

(defun parse-ecocyc-test ()
  
  (parse-ecocyc "/homedir/pkarp/h4/aic/ecocyc/misc-project-info/flatfiles/ecocyc-genes.dat"))

;;;========================================================================
;;; Prosite parsing function

(defun parse-prosite (filename 
		      &key 
		      index-attributes)
  (let ((separator-alist
	 '((ID (#\; #\.))
	   (AC (#\;))
	   (DT (#\; #\.))
	   (RU (#\: #\.))
	   (DE (#\.))
	   (PA (#\.))
	   (CC (#\; #\Newline))
	   (DO (#\;))
	   (NR (#\;))
	   (DR (#\;))
	   (MA (#\;))
	   (3D (#\;))
	   ))
	(continuator-alist
	 '((CC #\Newline)
	   ))
	removal-alist
	(data-spacers '(#\Space #\Tab))
	)
    
  
    (parse-flatfile-db filename "Prosite" "//"
		       separator-alist continuator-alist
		       removal-alist data-spacers
		       :comment-tag 'CC
		       :primary-key 'ID
		       :index-attributes index-attributes)))

;;; TEST VERSION

(defun parse-prosite-test ()
  
  (parse-prosite "/homedir/promero/franz/db-prosite.dat"))


;;;========================================================================
;;; Swissprot parsing function 

(defun parse-swissprot (filename 
		      &key 
		      index-attributes)
  (let ((separator-alist
	 '((ID (#\; #\. #\Space))
	   (AC (#\;))
	   (DT (#\Newline))
	   (OS (#\.))
	   (OC (#\; #\.))
	   (RC (#\;))
	   (RX (#\; #\.))
	   (RA (#\; #\,))
	   (RT (#\;))
	   (CC (#\;))
	   (DR (#\; #\.))
	   (KW (#\; #\.))
	   (FT (#\Newline))
	   (SQ (#\;))
	   ))
	(continuator-alist
	 '((RT #\Space)
	   (CC #\Newline)
	   ))
	(removal-alist
	 '((RT (#\"))
	   (SQ (#\Space))
	   ))
	(data-spacers '(#\Space #\Tab))
	)
    
  
    (parse-flatfile-db filename "SwissProt" "//"
		       separator-alist continuator-alist
		       removal-alist data-spacers
		       ;;:debug t
		       :comment-tag 'CC
		       :primary-key 'ID
		       :index-attributes index-attributes)))

;;; TEST VERSION

(defun parse-swissprot-test ()
    (parse-swissprot "/homedir/promero/franz/db-swiss-prot.dat"))
(defun parse-swissprot-test2 ()
    (parse-swissprot "/homedir/promero/franz/swiss_new.dat"))
(defun parse-swissprot-all ()
  (parse-swissprot "~pkarp/biodbs/swiss/39.0/sprot39.dat"))


;;;========================================================================
;;; Enzyme parsing function

(defun parse-enzyme (filename 
		     &key 
		     index-attributes)
  
  (let ((separator-alist
	 '((DE (#\.))
	   (AN (#\.))
	   (CA (#\.))
	   (CF (#\.))
	   (PR (#\;))
	   (DR (#\;))
	   (DI (#\; #\.))
	   ))
	(continuator-alist
	 '((CC #\Newline)
	   ))
	removal-alist
	(data-spacers '(#\Space #\Tab))
	)
    
    ;; Specialized call to parse-flatfile-db.
    
    (parse-flatfile-db filename "Enzyme" "//"
		       separator-alist continuator-alist
		       removal-alist data-spacers
		       :comment-tag 'CC
		       :primary-key 'ID
		       :index-attributes index-attributes)))

;;; TEST VERSION

(defun parse-enzyme-test ()
  (parse-enzyme "/homedir/promero/franz/db-enzyme.dat"))


;;;========================================================================
;;; General flatfile parsing function (parameterized).

(defun parse-flatfile-db (filename
			  name
			  entry-separator
			  separator-alist
			  continuator-alist
			  removal-alist
			  data-spacers
			  &key
			  (comment-tag nil)
			  primary-key
			  index-attributes
			  (debug nil))
  
  (let (input db temp-db-header)
    (format t "~%Loading ~A~%on ~A~%~%" filename (now))
    
    (biodb-gc)
    
    ;; Open DB file.
    (setq input (lisputils::open-file-for-block-reads filename))

    ;; Create db structure.
    (setq db 
      (make-biodb
       :name name
       :source-filename filename
       :primary-key primary-key
       :primary-index (make-hash-table :size 10000 :test #'equalp)
       :attribute-indices (loop for x in index-attributes
			      collect (list x (make-hash-table :size 10000 
							       :test #'equalp))
				      )))
    
    ;; Loop on DB file on a entry-by-entry basis. "db" is modified directly
    ;; by get-entries-from-file.
    (get-entries-from-file input 
			   db
			   entry-separator
			   comment-tag
			   data-spacers
			   separator-alist
			   continuator-alist
			   removal-alist
			   :debug debug)
    (close input)

    (when debug
      (format t "Number of entries loaded: ~D~%" 
	      (length (biodb-entry-list db))))
    
    ;; When db-header exists, combine all header lines into a single string.
    (when (biodb-db-header db)
      (if (= (length (biodb-db-header db)) 1)
	  (setf (biodb-db-header db) (first (biodb-db-header db)))
	(progn 
	  (loop for str in (biodb-db-header db)
	      if temp-db-header
	      do (setq temp-db-header
		   (concatenate 'string temp-db-header
			        (string #\Newline) str))
	      else
	      do (setq temp-db-header str))
	  (setf (biodb-db-header db) temp-db-header))))

    ;; Return database structure.
    db
    ) )

;;;========================================================================

(defun get-entries-from-file (input db entry-separator comment-tag data-spacers
			      separator-alist continuator-alist removal-alist
			      &key (debug nil))
  (let (positions			; list of buffer positions for a
					; tag group (tag + data)
	entry 
	header-found?
	(buffer-size 50000)
	tag-groups
	)
    (declare (special block-string buffer buffer-tail))
    (setq buffer (cl:make-array buffer-size :allocation :old
			     :element-type 'character))
    (setq buffer-tail 0)
    (setq tag-groups 0)
    
    (when debug
      (format t "~%Entering get-entries-from-file~%~%"))
    
    ;; Assemble entry list.
    (loop for n from 1
        do (setq buffer-tail 0)
	   ;; Copy entry from file to a buffer. Pointers for each field in
	   ;; entry are saved in "positions".
	   (setq positions 
	     (put-entry-in-buffer input entry-separator data-spacers))
	   (setq tag-groups (+ tag-groups (length positions)))
	   ;; Now buffer contains all entries, and entry-positions have
	   ;; the corresponding pointers to positions within the buffer 
	   ;; for each tag group within each entry.
	while (> buffer-tail 0)
	      ;; Work until there is something in buffer.
	      
	do (when (and debug
		      (= (rem n 1000) 0))
	     (format t "~A - ~D entries...~%" (now) n))
					; Print # of entries on screen every 
					; 1,000 entries (when debugging).
	   (when (and debug
		      (= (rem n 5000) 0))
		      (format t "~%~A (~D entries)~%" (now) n)
		      (cl:room t)
		      (format t "~%Newspace:~%~%")
		      (excl:print-type-counts :new)
		      (format t "~%Oldspace:~%~%")
		      (excl:print-type-counts :old)
		      (terpri))
					; Print time and garbage collection
					; parameters on screen every 
					; 5,000 entries (when debugging).
	   (setq entry (get-entry-from-buffer positions
					      separator-alist
					      continuator-alist 
					      removal-alist))
	   (when debug
	     (format t "Entry ~D:~%~S~%" n entry))
	   ;; Check if entry correspond to a DB header. If so, save the
	   ;; header as a single string in "db-header" (header will not
	   ;; go into entry-list).
	   (cond ((and comment-tag
		       (= (length entry) 1)
		       (equal (caar entry) comment-tag)
		       (not header-found?))
		  (setf (biodb-db-header db) (cdar entry))
		  (setq header-found? t)
		  )
		 (t
		  (push entry (biodb-entry-list db))
		  (when (biodb-primary-key db)
		    (index-entry 
		     entry (biodb-primary-key db)
		     :attribute-indices (biodb-attribute-indices db)
		     :db db) )
		  ))
	   )
    (when debug
      (format t "Tag-groups = ~D~%~%" tag-groups)
      (format t "Entry list:~%~S~%" (biodb-entry-list db))) ; debug
    t
    ))

;;;========================================================================
;;; Put entry from block-string into a buffer.

(defun put-entry-in-buffer (input entry-separator data-spacers)
  (let (positions
	block-start 
	old-block-start 
	block-end 
	old-block-end 
	block-tag-end
	block-data-start 
	old-block-tag-end 
	old-block-data-start 
	old-buffer-tail 
	buffer-tag-end
	buffer-data-start)
    
    (declare (special block-string buffer buffer-tail))
    ;; Read first line.
    
    (multiple-value-setq (block-string old-block-start old-block-end) 
      (lisputils::read-line-from-block-file input))
    
    ;; Unless no line were read...
    (when old-block-start
      (setq old-buffer-tail buffer-tail)
      (setq buffer-tail (put-chars-in-buffer buffer block-string
					     :start old-block-start 
					     :end old-block-end
					     :buffer-start buffer-tail))
      (multiple-value-setq (old-block-tag-end old-block-data-start)
	(get-tag-position old-block-start (- old-block-end 1) 
			  data-spacers block-string))
      (multiple-value-setq (buffer-tag-end buffer-data-start)
	(get-tag-position old-buffer-tail (- buffer-tail 1) 
			  data-spacers buffer))
      (loop
	  do (multiple-value-setq (block-string block-start block-end) 
	       (lisputils::read-line-from-block-file input))
	  while block-start
	  until (entry-separator-p block-start entry-separator block-string)
		
	       ;;; Check tags. If tags are equal, put new line minus tag in
	       ;;; buffer. If different, push the end position of the previous
	       ;;; ("old") line into "positions" and write complete new line
	       ;;; on buffer.
	  do (multiple-value-setq (block-tag-end block-data-start)
	       (get-tag-position block-start (- block-end 1) 
				 data-spacers block-string))
	     (cond ((or (< block-tag-end block-start)
			(same-tag-p block-string block-start block-tag-end 
				    :start2 old-block-start
				    :tag-end2 old-block-tag-end))
		    (setq buffer-tail (put-chars-in-buffer 
				       buffer block-string 
				       :char #\Newline
				       :buffer-start buffer-tail))
		    (setq buffer-tail (put-chars-in-buffer 
				       buffer block-string 
				       :start block-data-start 
				       :end block-end 
				       :buffer-start buffer-tail))
		    )
		   (t
		    (push (list buffer-tag-end buffer-data-start 
				buffer-tail)
			  positions)	; End of last line in buffer
		    (setq old-buffer-tail buffer-tail)
		    (setq buffer-tail (put-chars-in-buffer 
				       buffer block-string 
				       :start block-start :end block-end
				       :buffer-start buffer-tail))
		    (setq old-block-tag-end block-tag-end)
		    (setq old-block-data-start block-data-start)
		    (setq old-block-start block-start)
		    (multiple-value-setq (buffer-tag-end 
					  buffer-data-start)
		      (get-tag-position (+ old-buffer-tail 1) 
					(- buffer-tail 1) 
					data-spacers buffer))
		    ))
	     )
      ;; entry-separator found. Push positions for last tag group
      ;; into positions, then add positions for this entry to
      ;; entry positions.
      (push (list buffer-tag-end buffer-data-start buffer-tail)
	    positions)
      
      (setq positions (reverse positions))
      )

    ))


;;;========================================================================
;;; Parse entry from buffer to create an alist to be placed into the 
;;; entry-listat the db structure.

(defun get-entry-from-buffer (positions separator-alist
			      continuator-alist removal-alist)
  (let (continuator 
	i
	old-end
	start
	tag
	separators
	separator-positions
	separators-spacers
	removal-chars
	value
	(spacers '(#\Space #\tab))
	entry)
    (declare (special buffer entry-list))
    
    (setq old-end 0)
    
    (loop for pos in positions
	for tag-end = (first pos)
	for data-start = (second pos)
	for end = (third pos)
	for values = nil
	for last-sep? = nil
	do (if (> old-end 0)
	       (setq start (+ old-end 1))
	     (setq start 0))
	   (setq tag (intern (subseq buffer start (+ tag-end 1))))
	   (setq continuator (cadr (assoc tag continuator-alist
					   :test #'equal)))
	   
	   (setq separators (or (cadr (assoc 'default separator-alist 
					     :test #'equal))
				(cadr (assoc tag separator-alist 
					     :test #'equal))))
	   (setq separators-spacers (append separators spacers))
	   (setq removal-chars (cadr (assoc tag removal-alist 
					    :test #'equal)))
	   (setq i data-start)
	   
	   (setq separator-positions
	      (loop for j from i to end
		  for char = (schar buffer j)
		  if (not last-sep?)
		  when (find char separators)
		  collect j 
		  and do (setq last-sep? t)
		  end
		  else do (unless (find char separators)
			    (setq last-sep? nil))))

	   (setq separator-positions
	     (if (not separator-positions)
		 (list (+ end 1))
	       (if (< (car (last separator-positions)) end)
		   (append separator-positions (list (+ end 1)))
		 separator-positions)))
	   
	   (loop for j in separator-positions
	       do (loop while 
			(and (<= i end)
			     (find (schar buffer i) 
				   separators-spacers
				   :test #'eq))
		      do (incf i))
		  (when (> j i)
		    (setq value 
		      (extract-string buffer i j :continuator continuator
				      :removal-chars removal-chars)))
		  (push value values)
		  (setq i j))
	   (push (cons tag (reverse values)) entry)
	   (setq old-end end)
	   )
    (reverse entry)
    ))


;;;========================================================================
;;; Compare the tags from two lines in block-string.

(defun same-tag-p (block-string start1 tag-end1 &key start2 tag-end2 string)
  (cond (string
	 (when (= (- tag-end1 start1) (- (length string) 1))
	   (loop for i from start1 to tag-end1
	       for j from 0
	       always (eq (schar block-string i)
			  (schar string j)))))
	(start2
	 (when (= (- tag-end1 start1) (- tag-end2 start2))
	   (loop for i from start1 to tag-end1
	       for j from start2
	       always (eq (schar block-string i)
			  (schar block-string j))))))
  )

;;;========================================================================
;;; Locate the tag's end and the start of data in input line.

(defun get-tag-position (start end data-spacers block)
  (let (tag-end-position data-start-position spacer-start)

    (if (consp data-spacers)
	(progn
	  (setq tag-end-position
	    (loop for i from start
		until (or (find (schar block i) data-spacers)
			  (> i end))
		finally (return (- i 1))))
	  
	  (setq data-start-position
	    (if (= tag-end-position end)
		end
	      (loop for i from (+ tag-end-position 1)
		  while (find (schar block i) data-spacers)
		  finally (return i))))
	  (values tag-end-position data-start-position))
      (progn
	(setq spacer-start
	  (loop for i from start to (- end (length data-spacers))
	      until (string= data-spacers block 
			     :start2 i :end2 (+ i (length data-spacers)))
	      finally (return i)))
	(setq tag-end-position (- spacer-start 1))
	(setq data-start-position (+ spacer-start (length data-spacers)))
	(values tag-end-position data-start-position)
	))
    ))

  
;;;========================================================================
;;; Check if line consists of an entry-separator.

(defun entry-separator-p (start entry-separator block-string)  
  (let (len)
    (setq len (length entry-separator))
    (loop for i from 0 to (- len 1)
	for j = (+ i start)
	always (eq (schar entry-separator i)
		   (schar block-string j)))
    ))
  
;;;========================================================================
;;; Put characters in buffer. 
;;;
;;; Keys:
;;;
;;; char           if non-NIL, put character in "char" in buffer.
;;; line           if non-NIL, put string in "line" in buffer.
;;; start & end    if non-NIL, put the substring of "block-string"
;;;                delimited by "start" and "end" in buffer.
;;; buffer-start   starting position on buffer.

(defun put-chars-in-buffer (buffer block-string
			    &key char line start end (buffer-start 0))
  (unless (= buffer-start 0) (incf buffer-start))
  (cond (char
	 (setf (schar buffer buffer-start) char)
	 buffer-start)
	(line
	 (loop for i from 0 to (- (length line) 1)
	     for j from buffer-start
	     do (setf (schar buffer j) (schar line i))
	     finally (return j)))
	(start
	 (if (= (+ start 1) end)
	     (decf buffer-start)
	   (loop for i from start to (- end 1)
	       for j from buffer-start
	       do (setf (schar buffer j) (schar block-string i))
	       finally (return j)))
	 )))

;;;========================================================================


(defun extract-string (buffer start end &key continuator removal-chars)
  (let (newlines removed len string j)
    (decf end)
    ;; Count number of newline characters in the string to be extracted.
    (setq newlines (loop for i from start to end
		       when (eq #\Newline (schar buffer i))
		       count i))
    ;; Count number of characters to be removed from the original string
    ;; to be extracted.
    (when removal-chars
      (setq removed (loop for i from start to end
			when (find (schar buffer i) removal-chars)
			count i)))
    ;; Adjust length of the extracted string according to the number of
    ;; characters to be eliminated.
    (setq len (- (+ end 1) start))
    (unless continuator (setq len (- len newlines)))
    (when removed (setq len (- len removed)))
    
    ;; Create an appropriately sized string nad place it directly in
    ;; oldspace.
    (setq string (cl:make-array len :allocation :old :element-type 'character))
    (setq j 0)
    
    ;; Place characters from buffer in string, while handling newlines
    ;; and characters to be removed.
    (loop for i from start to end
	for char = (schar buffer i)
	if (eq char #\Newline)
	do (when continuator
	     (setf (schar string j) continuator)
	     (incf j))
	else
	unless (and removal-chars
		    (find char removal-chars))
	do (setf (schar string j) char)
	   (incf j))
    ;; Return string.
    string))


;;;=====================================================================
;;; Returns a string with current date and time as: "mm-dd-yyyy_hr:min:sec"

(defun now ()
  (let (sec min hr day month year)
    (multiple-value-setq (sec min hr day month year)
      (get-decoded-time))
    (concatenate 'string
      (prin1-to-string month) "-" (prin1-to-string day) "-" 
      (prin1-to-string year) "_" (prin1-to-string hr) ":" 
      (prin1-to-string min) ":" (prin1-to-string sec))))

;;;=====================================================================

(defun biodb-gc ()
  (setf (sys:gsgc-parameter :generation-spread) 4)
  (setq excl:*global-gc-behavior* '(100 ))
  (sys:resize-areas :new 10000000
		    :old 20000000
		    :sift-old-areas nil
		    :pack-heap nil)
  )


;;;=====================================================================

(defvar *current-db* nil
  "This variable defines a current DB that will be accessed by default
by this package.  Its value is a BioDB defstruct.")

;;;=====================================================================

(defun goto-db (db)
  (unless db
    (error "Database descriptor is NIL")
    )
  (unless (eq 'BioDB (type-of db))
    (error "Database descriptor must be a BioDB defstruct")
    )
  
  (setq *current-db* db)
  (or (biodb-name db) (biodb-source-filename db))
  )

;;;=====================================================================

(defun validate-db (db)
  (unless db
    (error "Database descriptor is NIL -- you may have neglected to load a BioDB or to
set the current DB using goto-db~%")
    )
  )

;;;=====================================================================
;; Given a list of entries, prints the value of the primary-key attribute
;; for each entry.

(defun print-ids (entry-list &key (db *current-db*))
  (let ((pkey (biodb-primary-key db)))
    (when pkey
      (loop for e in entry-list
	  do
	    (format t "~A~%" (get-value e pkey :db db))
	    ) )
    ) )


;;;=====================================================================
;; Returns all entries for BioDB

(defun get-all-entries (&optional (db *current-db*))
  (biodb-entry-list db)
  )

;;;=====================================================================
;; Returns the entry of DB whose primary key has the value Value.

(defun get-entry (value &key (db *current-db*))
  (validate-db db)
  (gethash value (biodb-primary-index db))
  )



;;;=====================================================================
;; Returns a list of the one or more entries of DB such that a value
;; of Attribute of that entry is Value.

(defun get-entries-by-indexed-attribute (attribute value &key (db *current-db*))
  (let (htable)
    (validate-db db)
    (setq htable (second (assoc attribute (biodb-attribute-indices db))))
    (if htable
	(gethash value htable)
      (error "Attribute ~A is not indexed for BioDB ~A~%" attribute db)
      )
    ) )



;;;=====================================================================
;; Returns a list of the values of Attribute of Entry.

(defun get-values (entry attribute &key (db *current-db*))
  (validate-db db)
  (cdr (assoc attribute entry))
  )


;;;=====================================================================
;; Returns the first value of Attribute of Entry.

(defun get-value (entry attribute &key (db *current-db*))
  (validate-db db)
  (second (assoc attribute entry))
  )


;;;=====================================================================
;; Replaces all existing values of Attribute of Entry with Values.

(defun put-values (entry attribute values &key (db *current-db*))
  (let (x)
    (validate-db db)

    ;; If the user has not passed us a list of values, listify it.
    (unless (consp values)
      (setq values (list values)) )

    (if (setq x (assoc attribute entry))
	(rplacd x values)
      (nconc entry (list (cons attribute values)))
      )
    ) )


;;;=====================================================================
;; Adds Value as an additional value of Attribute of Entry.

(defun add-value (entry attribute value &key (db *current-db*))
  (let (x)
    (validate-db db)
    (if (setq x (assoc attribute entry))
	(rplacd x (cons value (cdr x)))
      (nconc entry (list (list attribute value)))
      )
    ) )

;;;=====================================================================

(defun replace-value (entry attribute old-value new-value
		      &key 
		      (db *current-db*)
		      (test #'equal)
		      )
  (validate-db db)
  (let ((values (get-values entry attribute :db db)))
    (when (member old-value values :test test)
      (put-values entry attribute (substitute new-value old-value values 
					      :test test) )
      )
    ) )


;;;=====================================================================
;; Given a single Entry (an alist) in a BioDB, add new elements for that entry
;; to the primary-key index for this BioDB, and to any additional indices that
;; are defined for other attributes for this BioDB.

(defun index-entry (entry primary-key &key attribute-indices (db *current-db*))
  (let (v htable hashed-value)
    (validate-db db)

    (setq v (get-value entry primary-key :db db))
    (unless v 
      (format *error-output* "Warning: The following entry has no value for the primary 
attribute ~A, and therefore will not be indexed:~%~A~%"
	      primary-key entry)
      )

    (setq htable (biodb-primary-index db))
    (if (setq hashed-value (gethash v htable))
	(unless (eq hashed-value entry)
	  (format *error-output* "Warning: Duplicate DB entry under primary key ~A = ~A
will not be indexed.~%" 
		  primary-key v)
	  )
      (setf (gethash v htable) entry)
      )
    
    (loop for (attribute htable) in attribute-indices
	for values = (get-values entry attribute :db db)
	when htable
	do
	  (loop for v in values
	      do
		(setf (gethash v htable)
		  (cons entry (gethash v htable)) ) )
	  )
    ) )


