(in-package :lisputils)

(export '(memorize-file-as-list memorize-file-as-string))


;;  ***************************************************************
;;
;;               Copyright 2000 SRI International
;;
;;  ***************************************************************


;; ================================================================================

;; pkarp:Jun-28-2000 
;; Returns a list of strings, where each string contains a single line from
;; the file Filename, without embedded newlines.

(defun memorize-file-as-list (filename)
  (let (lines)
    (with-open-file (file filename
		     :direction :input)
      (loop for line = (read-line file nil nil)
	  while line
	  do
	    (push line lines)
	    ) )

    (nreverse lines)
    ) )

;; ================================================================================

;; pkarp:Jun-28-2000 
;; Returns a single large string containing the contents of the file Filename,
;; including embedded newlines.

(defun memorize-file-as-string (filename)
  (let (buffer)
    (with-open-file (file filename
		     :direction :input)
      (setq buffer (make-string (file-length file) :initial-element #\Space))
      (read-sequence buffer file)
      )
    
    buffer
    ) )


;; ================================================================================

;; The set of functions in this section provide more efficient reading 
;; of large files by Lisp.  The efficiency gain comes from reading the files
;; in large blocks into a fixed buffer, and by providing a line-at-a-time
;; processing facility that returns pointers into that buffer.  This approach
;; is more efficient than the normal Lisp read-line, which conses a new string
;; for every line it returns.

(defconstant +brbuf-len+ 102400)
(defvar *block-read-buffer* (make-string +brbuf-len+ :initial-element #\Space))
(defvar *buffer-start*)
(defvar *buffer-end* nil
  "The position after the last valid data character in the buffer.")
(defvar *buffer-stop* nil
  "The position after the last newline in the buffer.  Additional characters may
follow *buffer-stop*, with termination of data at *buffer-end*.")

;; Open Filename for reading via the block-reading system.  Note: only one file
;; may be open for block reading at a time using this facility.
;; Returns a normal Lisp stream connected to Filename.  To close the stream,
;; use the normal Lisp close operation.

(defun open-file-for-block-reads (filename)
  (let (file)
    (setq file (open filename :direction :input))
    (setq *buffer-end*    0
	  *buffer-stop*   0
	  *buffer-start*  0
	  )

    file
    ) )


;; Returns to the user a single line from the currently open block-reading file.
;; The user can call this function repeatedly to retrieve successive lines from the
;; file.  The function returns NIL for end-of-file.
;; Returns three values:
;;   o A string containing the line that is returned
;;   o Two integers Start and End, which are indices into the string of the start and
;;     end positions of the line that is returned.  End is the index into the string 
;;     of the newline character that ends the string.

;; Our internal processing depends on several globals.  *Buffer-start* is the index 
;; into *block-read-buffer* of the start of the next line; *buffer-stop* is the index 
;; of the last newline in the buffer, and *buffer-end* is the index of the end of 
;; the buffer (two different end indexes are needed because the buffer will not 
;; necessarily end on a newline boundary).

(defun read-line-from-block-file (stream)
  (let (stop start)

    ;; If the next line begins after the last newline in the buffer, we have to (a)
    ;; move the beginning of that next line to the start of the buffer, and (b) read
    ;; a new chunk of data into the remainder of the buffer.
    (when (>= *buffer-start* *buffer-stop*)
      (loop for i from 0
	  for j from *buffer-stop* below *buffer-end*
	  do
	    (setf (char *block-read-buffer* i) (char *block-read-buffer* j))
	    )
      
      (setq *buffer-end* (read-sequence *block-read-buffer* stream
					:start (- *buffer-end* *buffer-stop*))
	    *buffer-stop* *buffer-end*
	    *buffer-start* 0)
      ;; Check for end-of-file
      (when (= 0 *buffer-end*)
	(return-from read-line-from-block-file nil)
	) 
      
      ;; Set *buffer-stop* to point to the last newline.
      (loop for i from (1- *buffer-end*) downto 0
	  until (eql #\Newline (char *block-read-buffer* i))
	  do (decf *buffer-stop*)
	     )
      )

    ;; Return cursors to the start and end of the next line.
    (setq start *buffer-start*)
    (setq stop (position #\Newline *block-read-buffer* :start *buffer-start*))
    (setq *buffer-start* (1+ stop))

    (values *block-read-buffer* start stop)
    ) )

;; This function provides a simplified interface to the block-reading interface.
;; It returns the next line in the file as a newly consed string.  This interface
;; is easier to use, but less efficient than the function above.

(defun get-string-from-block-file (stream)
  (multiple-value-bind (buffer start stop)
      (read-line-from-block-file stream)
    (when buffer
      (subseq buffer start stop)
      )
    )
  )


(defun read-whole-file (filename &optional read-line?)
  (if read-line?
      (with-open-file (file filename :direction :input)
	(loop for line = (read-line file nil nil)
	    while line)
	)
    (let (file)
      (setq file (open-file-for-block-reads filename))
      (loop for line = (read-line-from-block-file file)
	  while line)
      )
    )
  )
